"""Tests suite for `aria2p`."""

from pathlib import Path

TESTS_DIR = Path(__file__).parent
TESTS_TMP_DIR = TESTS_DIR / "tmp"
TESTS_DATA_DIR = TESTS_DIR / "data"
CONFIGS_DIR = TESTS_DATA_DIR / "configs"
SESSIONS_DIR = TESTS_DATA_DIR / "sessions"

BUNSENLABS_TORRENT = TESTS_DATA_DIR / "bunsenlabs-helium-4.iso.torrent"
BUNSENLABS_MAGNET = "magnet:?xt=urn:btih:7fb1b254fdbdd8863d686c7fa61b3b0b671551b1&dn=bl-Helium-4-amd64.iso"
DEBIAN_METALINK = TESTS_DATA_DIR / "debian.metalink"
XUBUNTU_MIRRORS = [
    "http://ubuntutym2.u-toyama.ac.jp/xubuntu/18.04/release/xubuntu-18.04.1-desktop-amd64.iso",
    "http://ftp.free.fr/mirrors/ftp.xubuntu.com/releases/18.04/release/xubuntu-18.04.1-desktop-amd64.iso",
    "http://mirror.us.leaseweb.net/ubuntu-cdimage/xubuntu/releases/18.04/release/xubuntu-18.04.1-desktop-amd64.iso",
]

INPUT_FILES = [
    TESTS_DATA_DIR / "input_files" / "two-valid-downloads",
    TESTS_DATA_DIR / "input_files" / "one-valid-one-invalid-downloads",
    TESTS_DATA_DIR / "input_files" / "two-invalid-downloads",
]



"""Configuration for the pytest test suite."""

import json
import os
import random
import subprocess
import sys
import time
from pathlib import Path

import pytest
import requests

from aria2p import API, Client, enable_logger

from . import CONFIGS_DIR, SESSIONS_DIR


@pytest.fixture(autouse=True)
def tests_logs(request):
    # put logs in tests/logs
    log_path = Path("tests") / "logs"

    # tidy logs in subdirectories based on test module and class names
    module = request.module
    class_ = request.cls
    name = request.node.name + ".log"

    if module:
        log_path /= module.__name__.replace("tests.", "")
    if class_:
        log_path /= class_.__name__

    log_path.mkdir(parents=True, exist_ok=True)

    # append last part of the name and enable logger
    log_path /= name
    if log_path.exists():
        log_path.unlink()
    enable_logger(sink=log_path, level=os.environ.get("PYTEST_LOG_LEVEL", "TRACE"))


def spawn_and_wait_server(port=8779):
    process = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "uvicorn",
            "tests.http_server:app",
            "--port",
            str(port),
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    while True:
        try:
            requests.get(f"http://localhost:{port}/1024")
        except:
            time.sleep(0.1)
        else:
            break
    return process


@pytest.fixture(scope="session", autouse=True)
def http_server(tmp_path_factory, worker_id):
    if worker_id == "master":
        # single worker: just run the HTTP server
        process = spawn_and_wait_server()
        yield process
        process.kill()
        process.wait()
        return

    # get the temp directory shared by all workers
    root_tmp_dir = tmp_path_factory.getbasetemp().parent

    # try to get a lock
    lock = root_tmp_dir / "lock"
    try:
        lock.mkdir(exist_ok=False)
    except FileExistsError:
        yield  # failed, don't run the HTTP server
        return

    # got the lock, run the HTTP server
    process = spawn_and_wait_server()
    yield process
    process.kill()
    process.wait()


class _Aria2Server:
    def __init__(self, tmp_dir, port, config=None, session=None, secret=""):
        self.tmp_dir = tmp_dir
        self.port = port

        # create the command used to launch an aria2c process
        command = [
            "aria2c",
            f"--dir={self.tmp_dir}",
            "--file-allocation=none",
            "--quiet",
            "--enable-rpc=true",
            f"--rpc-listen-port={self.port}",
        ]
        if config:
            command.append(f"--conf-path={config}")
        else:
            # command.append("--no-conf")
            config = CONFIGS_DIR / "default.conf"
            command.append(f"--conf-path={config}")
        if session:
            if isinstance(session, list):
                session_path = self.tmp_dir / "_session.txt"
                with open(session_path, "w") as stream:
                    stream.write("\n".join(session))
                command.append(f"--input-file={session_path}")
            else:
                session_path = SESSIONS_DIR / session
                if not session_path.exists():
                    raise ValueError(f"no such session: {session}")
                command.append(f"--input-file={session_path}")
        if secret:
            command.append(f"--rpc-secret={secret}")

        self.command = command
        self.process = None

        # create the client with port
        self.client = Client(port=self.port, secret=secret, timeout=20)

        # create the API instance
        self.api = API(self.client)

    def start(self):
        while True:
            # create the subprocess
            self.process = subprocess.Popen(self.command)

            # make sure the server is running
            retries = 5
            while retries:
                try:
                    self.client.list_methods()
                except requests.ConnectionError:
                    time.sleep(0.1)
                    retries -= 1
                else:
                    break

            if retries:
                break

    def wait(self):
        while True:
            try:
                self.process.wait()
            except subprocess.TimeoutExpired:
                pass
            else:
                break

    def terminate(self):
        self.process.terminate()
        self.wait()

    def kill(self):
        self.process.kill()
        self.wait()

    def rmdir(self, directory=None):
        if directory is None:
            directory = self.tmp_dir
        for item in directory.iterdir():
            if item.is_dir():
                self.rmdir(item)
            else:
                item.unlink()
        directory.rmdir()

    def destroy(self, force=False):
        if force:
            self.kill()
        else:
            self.terminate()
        self.rmdir()


class Aria2Server:
    def __init__(self, tmp_dir, port, config=None, session=None, secret=""):
        self.server = _Aria2Server(tmp_dir, port, config, session, secret)

    def __enter__(self):
        self.server.start()
        return self.server

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.server.destroy(force=True)


ports_file = Path(".ports.json")


def get_lock():
    lockdir = Path(".lockdir")
    while True:
        try:
            lockdir.mkdir(exist_ok=False)
        except FileExistsError:
            time.sleep(0.025)
        else:
            break


def release_lock():
    Path(".lockdir").rmdir()


def get_random_port():
    return random.randint(15000, 16000)


def get_current_ports():
    try:
        return json.loads(ports_file.read_text())
    except FileNotFoundError:
        return []


def set_current_ports(ports):
    ports_file.write_text(json.dumps(ports))


def reserve_port():
    get_lock()

    ports = get_current_ports()
    port_number = get_random_port()
    while port_number in ports:
        port_number = get_random_port()
    ports.append(port_number)
    set_current_ports(ports)

    release_lock()
    return port_number


def release_port(port_number):
    get_lock()
    ports = get_current_ports()
    ports.remove(port_number)
    set_current_ports(ports)
    release_lock()


@pytest.fixture
def port():
    port_number = reserve_port()
    yield port_number
    release_port(port_number)


@pytest.fixture
def server(tmp_path, port):
    with Aria2Server(tmp_path, port) as server:
        yield server



"""Tests for the `utils` module."""

from datetime import timedelta

import pytest

from aria2p.utils import bool_or_value, bool_to_str, human_readable_bytes, human_readable_timedelta


@pytest.mark.parametrize(
    ("args", "kwargs", "expected"),
    [
        ([0], {}, "0.00B"),
        ([0], {"digits": 0}, "0B"),
        ([0], {"digits": 0, "delim": " "}, "0 B"),
        ([0], {"digits": 0, "delim": " ", "postfix": "/s"}, "0 B/s"),
        ([1024], {"digits": 0, "delim": " ", "postfix": "/s"}, "1 KiB/s"),
        ([2048], {}, "2.00KiB"),
        ([2048 * 1024], {}, "2.00MiB"),
        ([2048 * 1024 * 1024], {}, "2.00GiB"),
        ([2048 * 1024 * 1024 * 1024], {}, "2.00TiB"),
    ],
)
def test_human_readable_bytes(args, kwargs, expected):
    """
    Test the `human_readable_bytes` function.

    Arguments:
        args: Positional arguments passed to the function.
        kwargs: Keyword arguments passed to the function.
        expected: The expected result.
    """
    assert human_readable_bytes(*args, **kwargs) == expected


@pytest.mark.parametrize(
    ("td", "expected"),
    [
        (timedelta(seconds=0), "0s"),
        (timedelta(seconds=1), "1s"),
        (timedelta(seconds=75), "1m15s"),
        (timedelta(seconds=3601), "1h1s"),
        (timedelta(seconds=3600 * 24 + 2), "1d2s"),
        (timedelta(days=1, seconds=3600 * 24 + 2), "2d2s"),
        (timedelta(days=3, seconds=3), "3d3s"),
        (timedelta(days=3), "3d"),
        (timedelta(seconds=3600 * 3), "3h"),
        (timedelta(seconds=60 * 3), "3m"),
    ],
)
def test_human_readable_timedelta_force_print_0_seconds(td, expected):
    """
    Test the `human_readable_timedelta` function.

    Arguments:
        td: A timedelta.
        expected: The expected result.
    """
    assert human_readable_timedelta(td) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("true", True),
        ("false", False),
        ("mem", "mem"),
        (None, None),
        (0, 0),
        (1, 1),
    ],
)
def test_bool_or_value_true_is_true(value, expected):
    """
    Test the `bool_or_value` function.

    Arguments:
        value: Value passed to the function.
        expected: The expected result.
    """
    assert bool_or_value(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (True, "true"),
        (False, "false"),
        (None, None),
        (0, 0),
        (1, 1),
    ],
)
def test_bool_to_str_true_gives_true(value, expected):
    """
    Test the `bool_to_str` function.

    Arguments:
        value: Value passed to the function.
        expected: The expected result.
    """
    assert bool_to_str(value) == expected



"""Command to add magnets."""

import sys
from typing import List

from aria2p.api import API
from aria2p.utils import read_lines


def add_magnets(
    api: API,
    uris: List[str] = None,
    from_file: str = None,
    options: dict = None,
    position: int = None,
) -> int:
    """
    Add magnet subcommand.

    Arguments:
        api: The API instance to use.
        uris: The URIs of the magnets.
        from_file: Path to the file to read uris from.
        options: String of aria2c options to add to download.
        position: Position to add new download in the queue.

    Returns:
        int: Always 0.
    """
    ok = True

    if not uris:
        uris = []

    if from_file:
        try:
            uris.extend(read_lines(from_file))
        except OSError:
            print(f"Cannot open file: {from_file}", file=sys.stderr)
            ok = False

    for uri in uris:
        new_download = api.add_magnet(uri, options=options, position=position)
        print(f"Created download {new_download.gid}")

    return 0 if ok else 1



"""Tests for the `interface` module."""

import os
import sys
import time
from pathlib import Path

import pyperclip
import pytest
from asciimatics.event import KeyboardEvent, MouseEvent
from asciimatics.screen import Screen

from aria2p import interface as tui

from . import TESTS_DATA_DIR
from .conftest import Aria2Server

tui.Interface.frames = 20  # reduce tests time


class SpecialEvent:
    RESIZE = 1
    PASS_N_FRAMES = 2
    PASS_N_TICKS = 3
    RAISE = 4

    def __init__(self, event_type, value=None):
        self.type = event_type
        self.value = value


class Event:
    resize = SpecialEvent(SpecialEvent.RESIZE)
    pass_frame = SpecialEvent(SpecialEvent.PASS_N_FRAMES, 1)
    pass_tick = SpecialEvent(SpecialEvent.PASS_N_TICKS, 1)
    pass_half_tick = SpecialEvent(SpecialEvent.PASS_N_FRAMES, tui.Interface.frames / 2)
    pass_tick_and_a_half = SpecialEvent(SpecialEvent.PASS_N_FRAMES, tui.Interface.frames * 3 / 2)
    up = KeyboardEvent(Screen.KEY_UP)
    down = KeyboardEvent(Screen.KEY_DOWN)
    left = KeyboardEvent(Screen.KEY_LEFT)
    right = KeyboardEvent(Screen.KEY_RIGHT)
    delete = KeyboardEvent(Screen.KEY_DELETE)
    esc = KeyboardEvent(Screen.KEY_ESCAPE)
    f1 = KeyboardEvent(Screen.KEY_F1)
    f2 = KeyboardEvent(Screen.KEY_F2)
    f3 = KeyboardEvent(Screen.KEY_F3)
    f4 = KeyboardEvent(Screen.KEY_F4)
    f5 = KeyboardEvent(Screen.KEY_F5)
    f6 = KeyboardEvent(Screen.KEY_F6)
    f7 = KeyboardEvent(Screen.KEY_F7)
    f8 = KeyboardEvent(Screen.KEY_F8)
    f9 = KeyboardEvent(Screen.KEY_F9)
    f10 = KeyboardEvent(Screen.KEY_F10)
    f11 = KeyboardEvent(Screen.KEY_F11)
    f12 = KeyboardEvent(Screen.KEY_F12)
    enter = KeyboardEvent(ord("\n"))
    space = KeyboardEvent(ord(" "))

    @staticmethod
    def hit(letter):
        return KeyboardEvent(ord(letter))

    @staticmethod
    def exc(value):
        return SpecialEvent(SpecialEvent.RAISE, value)

    @staticmethod
    def pass_frames(value):
        return SpecialEvent(SpecialEvent.PASS_N_FRAMES, value)

    @staticmethod
    def pass_ticks(value):
        return SpecialEvent(SpecialEvent.PASS_N_TICKS, value)


def get_interface(patcher, api=None, events=None, append_q=True):
    if not events:
        events = []

    if append_q:
        events.append(KeyboardEvent(ord("q")))

    class MockedManagedScreen:
        def __enter__(self):
            return MockedScreen(events)

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    patcher.setattr(tui, "ManagedScreen", MockedManagedScreen)
    return tui.Interface(api=api)


def run_interface(patcher, api=None, events=None, append_q=True, **kwargs):
    interface = get_interface(patcher, api, events, append_q)
    for key, value in kwargs.items():
        setattr(interface, key, value)
    interface.run()
    return interface


class MockedScreen:
    def __init__(self, events):
        self.events = events
        self._has_resized = False
        self._pass_n_frames = 0
        self.print_at_calls = []
        self.paint_calls = []
        self.n_refresh = 0

    @property
    def dimensions(self):
        return 30, 80

    def has_resized(self):
        return self._has_resized

    def open(self):
        pass

    def close(self):
        pass

    def get_event(self):
        if self._pass_n_frames > 0:
            self._pass_n_frames -= 1
            return None
        event = self.events.pop(0)
        if isinstance(event, (KeyboardEvent, MouseEvent)):
            return event
        elif isinstance(event, SpecialEvent):
            if event.type == SpecialEvent.RESIZE:
                self._has_resized = True
            elif event.type == SpecialEvent.PASS_N_FRAMES:
                # we remove 1 because this event itself eats a frame
                self._pass_n_frames = event.value - 1
            elif event.type == SpecialEvent.PASS_N_TICKS:
                # we remove 1 because this event itself eats a frame
                self._pass_n_frames = (event.value * tui.Interface.frames) - 1
            elif event.type == SpecialEvent.RAISE:
                raise event.value
        return None

    def print_at(self, *args, **kwargs):
        if args[0].strip():
            self.print_at_calls.append({"args": args, "kwargs": kwargs})

    def paint(self, *args, **kwargs):
        if args[0].strip():
            self.paint_calls.append({"args": args, "kwargs": kwargs})

    def refresh(self):
        self.n_refresh += 1


def test_run(monkeypatch):
    interface = run_interface(monkeypatch)
    assert interface.screen
    assert interface.height == 30
    assert interface.width == 80


def test_resize(server, monkeypatch):
    interface = run_interface(monkeypatch, server.api, events=[Event.resize])
    # assert screen was renewed
    assert not interface.screen.has_resized()


def test_frames_plus_n(server, monkeypatch):
    n = 10
    interface = run_interface(monkeypatch, server.api, events=[Event.pass_frames(tui.Interface.frames + n)])
    assert interface.frame == n


def test_change_sort(server, monkeypatch):
    interface = run_interface(
        monkeypatch,
        server.api,
        events=[
            Event.hit(">"),
            Event.pass_half_tick,
            Event.hit("<"),
            Event.pass_half_tick,
            Event.hit("<"),
            Event.pass_tick,
        ],
    )
    assert interface.sort == tui.Interface.sort - 1


def test_move_focus(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        interface = run_interface(
            monkeypatch,
            server.api,
            events=[Event.up, Event.pass_tick] + [Event.down] * 30 + [Event.pass_tick] + [Event.up] * 5,
        )
    assert interface.focused == 0


def test_show_help(server, monkeypatch):
    interface = run_interface(monkeypatch, server.api, events=[Event.f1, Event.pass_tick, Event.enter])
    assert interface.screen.print_at_calls[-1]["args"][0].startswith("Press any key to return.")


def test_horizontal_scrolling(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="3-magnets.txt") as server:
        run_interface(
            monkeypatch,
            server.api,
            events=[Event.left]
            + [Event.pass_tick]
            + [Event.right] * 2
            + [Event.pass_tick_and_a_half]
            + [Event.right] * 20
            + [Event.pass_tick]
            + [Event.left] * 5,
        )


def test_log_exception(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        interface = get_interface(monkeypatch, server.api, events=[Event.exc(LookupError("some message"))])
        assert not interface.run()
    with open(Path("tests") / "logs" / "test_interface" / "test_log_exception.log") as log_file:
        lines = log_file.readlines()
    first_line = None
    for line in lines:
        if "ERROR" in line:
            first_line = line
            break
    last_line = lines[-1]
    assert "some message" in first_line
    assert "LookupError" in last_line


def test_select_sort(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="3-magnets.txt") as server:
        run_interface(
            monkeypatch,
            server.api,
            events=[
                Event.f6,
                Event.pass_tick_and_a_half,
                Event.down,
                Event.pass_tick,
                Event.down,
                Event.pass_tick,
                Event.enter,
                Event.pass_tick,
            ],
        )


def test_mouse_event(tmp_path, port, monkeypatch):
    reverse = tui.Interface.reverse
    with Aria2Server(tmp_path, port, session="3-magnets.txt") as server:
        interface = run_interface(
            monkeypatch,
            server.api,
            events=[MouseEvent(x=tui.Interface.x_offset, y=tui.Interface.y_offset, buttons=MouseEvent.LEFT_CLICK)] * 2,
        )
    assert interface.sort == 0
    assert interface.reverse is not reverse


def test_vertical_scrolling(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="50-dls.txt") as server:
        run_interface(monkeypatch, server.api, events=[Event.pass_frame] + [Event.down] * 40 + [Event.up] * 30)


def test_follow_focused(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="3-magnets.txt") as server:
        interface = run_interface(
            monkeypatch,
            server.api,
            events=[Event.hit("<"), Event.hit("<"), Event.pass_tick, Event.hit("F"), Event.hit("I"), Event.pass_tick],
        )
    assert interface.sort == 0
    assert interface.focused == 2


def test_remove_ask(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="very-small-download.txt") as server:
        download = server.api.get_downloads()[0]
        while not download.live.is_complete:
            if download.has_failed:
                pytest.xfail("Failed to establish connection (sporadic error)")
            time.sleep(0.1)
        assert download.root_files_paths[0].exists()
        interface = run_interface(
            monkeypatch,
            server.api,
            events=[Event.pass_frame, Event.delete, Event.pass_tick, Event.down, Event.enter, Event.pass_tick],
        )
        assert not download.root_files_paths[0].exists()
        assert interface.last_remove_choice == 1


def test_setup(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        run_interface(monkeypatch, server.api, events=[Event.pass_frame, Event.f2, Event.pass_tick])


def test_toggle_resume_pause(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="big-download.txt") as server:
        interface = run_interface(
            monkeypatch,
            server.api,
            events=[
                Event.pass_frame,
                Event.space,
                Event.pass_tick,
                Event.pass_tick,
                Event.space,
                Event.pass_tick,
                Event.pass_tick,
                Event.space,
            ],
        )

        time.sleep(1)
        interface.data[0].update()
    if interface.data[0].has_failed:
        pytest.xfail("Failed to establish connection (sporadic error)")
    assert interface.data[0].is_paused


def test_priority(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        run_interface(
            monkeypatch,
            server.api,
            events=[
                Event.pass_frame,
                Event.hit("["),
                Event.pass_tick,
                Event.hit("]"),
                Event.pass_tick,
                Event.hit("]"),
                Event.pass_tick,
            ],
            sort=7,
        )


def test_sort_edges(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        run_interface(monkeypatch, server.api, events=[Event.hit("<")], sort=0)
        run_interface(monkeypatch, server.api, events=[Event.hit(">")], sort=7)


def test_remember_last_remove(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        interface = run_interface(
            monkeypatch,
            server.api,
            events=[Event.pass_frame, Event.delete, Event.down, Event.enter, Event.delete, Event.esc],
        )
    assert interface.last_remove_choice == 1
    assert interface.side_focused == 1


def test_autoclear(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="very-small-download.txt") as server:
        download = server.api.get_downloads()[0]
        while not download.live.is_complete:
            if download.has_failed:
                pytest.xfail("Failed to establish connection (sporadic error)")
            time.sleep(0.1)
        interface = run_interface(monkeypatch, server.api, events=[Event.pass_frame, Event.hit("c"), Event.pass_tick])
    assert not interface.data


def test_side_column_edges(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        run_interface(
            monkeypatch,
            server.api,
            events=[
                Event.pass_frame,
                Event.delete,
                Event.down,
                Event.up,
                Event.up,
                Event.down,
                Event.down,
                Event.down,
                Event.down,
                Event.esc,
            ],
        )
        run_interface(
            monkeypatch,
            server.api,
            events=[Event.pass_frame, Event.f6]
            + [Event.up] * len(tui.Interface.columns_order)
            + [Event.down] * len(tui.Interface.columns_order)
            + [Event.esc],
        )


def test_click_row(tmp_path, port, monkeypatch):
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        interface = run_interface(
            monkeypatch, server.api, events=[Event.pass_frame, MouseEvent(x=10, y=2, buttons=MouseEvent.LEFT_CLICK)]
        )
    assert interface.focused == 1


def test_click_out_bounds(server, monkeypatch):
    run_interface(
        monkeypatch, server.api, events=[Event.pass_frame, MouseEvent(x=1000, y=0, buttons=MouseEvent.LEFT_CLICK)]
    )
    with open(Path("tests") / "logs" / "test_interface" / "test_click_out_bounds.log") as log_file:
        lines = log_file.readlines()
    error_line = None
    for line in lines:
        if "ERROR" in line:
            error_line = line
            break
    assert error_line
    assert "clicked outside of boundaries" in error_line


@pytest.mark.skipif(
    sys.platform != "win32" and "CI" in os.environ,
    reason="Can't get pyperclip to work on Linux or MacOS in CI",
)
def test_add_downloads_uris(server, monkeypatch):
    clipboard_selection_downloads = ""
    primary_selection_downloads = ""

    uri1 = "http://localhost:8779/1"
    magnet1 = "magnet:?xt=urn:btih:RX46NCATYQRS3MCQNSEXVZGCCDNKTASQ"
    clipboard_selection_downloads += "\n".join([uri1, magnet1])

    uri2 = "http://localhost:8779/2"
    magnet2 = "magnet:?xt=urn:btih:VLYICEBJDQQ64SUGREZHD4IAD2FVCJCS"
    primary_selection_downloads += "\n".join([uri2, magnet2])

    # clipboard selection
    pyperclip.copy(clipboard_selection_downloads)

    # primary selection
    pyperclip.copy(primary_selection_downloads, primary=True)

    interface = run_interface(
        monkeypatch,
        server.api,
        events=[
            Event.pass_frame,
            Event.hit("a"),
            Event.esc,
            Event.hit("a"),
            Event.down,
            Event.enter,
            Event.enter,
            Event.esc,
            Event.pass_tick_and_a_half,
        ],
    )
    # clear clipboards
    pyperclip.copy("")
    pyperclip.copy("", primary=True)
    assert len(interface.data) == 2


@pytest.mark.skipif(
    sys.platform != "win32" and "CI" in os.environ,
    reason="Can't get pyperclip to work on Linux or MacOS in CI",
)
def test_add_downloads_torrents_and_metalinks(server, monkeypatch):
    torrent_file = TESTS_DATA_DIR / "bunsenlabs-helium-4.iso.torrent"
    metalink_file = TESTS_DATA_DIR / "debian.metalink"

    clipboard_selection_download = f"{torrent_file.absolute()}"
    primary_selection_download = f"{metalink_file.absolute()}"

    # clipboard selection
    pyperclip.copy(clipboard_selection_download)

    # primary selection
    pyperclip.copy(primary_selection_download, primary=True)

    interface = run_interface(
        monkeypatch,
        server.api,
        events=[
            Event.pass_frame,
            Event.hit("a"),
            Event.esc,
            Event.pass_tick_and_a_half,
            Event.hit("a"),
            Event.down,
            Event.up,
            Event.enter,
            Event.enter,
            Event.pass_tick_and_a_half,
        ],
    )
    pyperclip.copy("")
    pyperclip.copy("", primary=True)
    assert len(interface.data) == 2



"""Command to show downloads."""

from aria2p.api import API


def show(api: API) -> int:
    """
    Show subcommand.

    Arguments:
        api: The API instance to use.

    Returns:
        int: Always 0.
    """
    downloads = api.get_downloads()

    def print_line(*args):  # noqa: WPS430 (nested function)
        print("{:<17} {:<9} {:>8} {:>12} {:>12} {:>8}  {}".format(*args))  # noqa: P101 (unindexed params)

    print_line("GID", "STATUS", "PROGRESS", "DOWN_SPEED", "UP_SPEED", "ETA", "NAME")

    for download in downloads:
        print_line(
            download.gid,
            download.status,
            download.progress_string(),
            download.download_speed_string(),
            download.upload_speed_string(),
            download.eta_string(),
            download.name,
        )

    return 0



from fastapi import FastAPI
from fastapi.responses import StreamingResponse


def translate_size(size):
    try:
        return int(size)
    except ValueError:
        pass
    size = size.lower()
    if size.endswith("k"):
        multiplier = 2 ** 10
    elif size.endswith("m"):
        multiplier = 2 ** 20
    elif size.endswith("g"):
        multiplier = 2 ** 30
    else:
        raise ValueError("size unit not supported:", size)
    return int(size.rstrip("kmg")) * multiplier


async def virtual_file(size, chunks=4096):
    while size > 0:
        yield b"1" * min(size, chunks)
        size -= chunks


app = FastAPI()


@app.get("/{size}")
async def get(size):
    return StreamingResponse(
        virtual_file(translate_size(size)),
        media_type="application/octet-stream",
    )



"""Command to listen to notifications from the server."""

import sys
from importlib import util as importlib_util
from pathlib import Path
from typing import List

from aria2p.api import API
from aria2p.types import PathOrStr


def listen(
    api: API,
    callbacks_module: PathOrStr = None,
    event_types: List[str] = None,
    timeout: int = 5,
) -> int:
    """
    Listen subcommand.

    Arguments:
        api: The API instance to use.
        callbacks_module: The path to the module to import, containing the callbacks as functions.
        event_types: The event types to process.
        timeout: The timeout to pass to the WebSocket connection, in seconds.

    Returns:
        int: Always 0.
    """
    if not callbacks_module:
        print("aria2p: listen: Please provide the callback module file path with -c option", file=sys.stderr)
        return 1

    if isinstance(callbacks_module, Path):
        callbacks_module = str(callbacks_module)

    if not event_types:
        event_types = ["start", "pause", "stop", "error", "complete", "btcomplete"]

    spec = importlib_util.spec_from_file_location("aria2p_callbacks", callbacks_module)

    if spec is None:
        print(f"aria2p: Could not import module file {callbacks_module}", file=sys.stderr)
        return 1

    callbacks = importlib_util.module_from_spec(spec)

    if callbacks is None:
        print(f"aria2p: Could not import module file {callbacks_module}", file=sys.stderr)
        return 1

    spec.loader.exec_module(callbacks)  # type: ignore

    callbacks_kwargs = {}
    for callback_name in (  # noqa: WPS352 (multiline loop)
        "on_download_start",
        "on_download_pause",
        "on_download_stop",
        "on_download_error",
        "on_download_complete",
        "on_bt_download_complete",
    ):
        if callback_name[3:].replace("download", "").replace("_", "") in event_types:
            callback = getattr(callbacks, callback_name, None)
            if callback:
                callbacks_kwargs[callback_name] = callback

    api.listen_to_notifications(timeout=timeout, handle_signals=True, threaded=False, **callbacks_kwargs)
    return 0



"""Tests for the `client` module."""

import json
import os
import signal
import threading
import time
from base64 import b64encode
from copy import deepcopy

import pytest
import requests
from responses import mock as responses

from aria2p import Client, ClientException
from aria2p.client import JSONRPC_CODES, JSONRPC_PARSER_ERROR, Notification

from . import BUNSENLABS_MAGNET, BUNSENLABS_TORRENT, CONFIGS_DIR, DEBIAN_METALINK, SESSIONS_DIR, XUBUNTU_MIRRORS
from .conftest import Aria2Server


class TestParameters:
    # callback that return params of a single call as result
    @staticmethod
    def call_params_callback(request):
        payload = json.loads(request.body)
        resp_body = {"result": payload["params"]}
        return 200, {}, json.dumps(resp_body)

    # callback that return params of a batch call as result
    @staticmethod
    def batch_call_params_callback(request):
        payload = json.loads(request.body)
        resp_body = [{"result": method["params"]} for method in payload]
        return 200, {}, json.dumps(resp_body)

    @responses.activate
    def test_insert_secret_with_aria2_method_call(self):
        # create client with secret
        secret = "hello"
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params = ["param1", "param2"]
        # copy params and insert secret
        expected_params = deepcopy(params)
        expected_params.insert(0, f"token:{secret}")

        # call function and assert result
        resp = client.call(client.ADD_URI, params, insert_secret=True)
        assert resp == expected_params

    @responses.activate
    def test_insert_secret_with_system_multicall(self):
        # create client with secret
        secret = "hello"
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params = [
            [
                {"methodName": client.ADD_URI, "params": ["param1", "param2"]},
                {"methodName": client.ADD_URI, "params": ["param3", "param4"]},
            ]
        ]
        # copy params and insert secret
        expected_params = deepcopy(params)
        for param in expected_params[0]:
            param["params"].insert(0, f"token:{secret}")

        # call function and assert result
        resp = client.call(client.MULTICALL, params, insert_secret=True)
        assert resp == expected_params

    @responses.activate
    def test_does_not_insert_secret_with_unknown_method_call(self):
        # create client with secret
        secret = "hello"
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params = ["param1", "param2"]

        # call function and assert result
        resp = client.call("other.method", params, insert_secret=True)
        assert secret not in resp

    @responses.activate
    def test_does_not_insert_secret_if_told_so(self):
        # create client with secret
        secret = "hello"
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params = ["param1", "param2"]

        # call function and assert result
        resp = client.call("other.method", params, insert_secret=False)
        assert secret not in resp

    def test_client_str_returns_client_server(self):
        host = "https://localhost:8779/"
        port = 7100
        client = Client(host, port)
        assert client.server == f"{host.rstrip('/')}:{port}/jsonrpc" == str(client)

    @responses.activate
    def test_batch_call(self):
        client = Client()

        responses.add_callback(responses.POST, client.server, callback=self.batch_call_params_callback)

        # create params
        params_1 = ["param1", "param2"]
        params_2 = ["param3", "param4"]
        # copy params and insert secret
        expected_params = [params_1, params_2]

        # call function and assert result
        resp = client.batch_call([(client.ADD_URI, params_1, 0), (client.ADD_METALINK, params_2, 1)])
        assert resp == expected_params

    @responses.activate
    def test_insert_secret_with_batch_call(self):
        # create client with secret
        secret = "hello"
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.batch_call_params_callback)

        # create params
        params_1 = ["param1", "param2"]
        params_2 = ["param3", "param4"]
        # copy params and insert secret
        expected_params = [deepcopy(params_1), deepcopy(params_2)]
        for p in expected_params:
            p.insert(0, f"token:{secret}")

        # call function and assert result
        resp = client.batch_call(
            [(client.ADD_URI, params_1, 0), (client.ADD_METALINK, params_2, 1)], insert_secret=True
        )
        assert resp == expected_params

    @responses.activate
    def test_multicall2(self):
        client = Client()

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params_1 = ["0000000000000001"]
        params_2 = ["2fa07b6e85c40205"]
        calls = [(client.REMOVE, params_1), (client.REMOVE, params_2)]
        # copy params and insert secret
        expected_params = [
            [
                {"methodName": client.REMOVE, "params": deepcopy(params_1)},
                {"methodName": client.REMOVE, "params": deepcopy(params_2)},
            ]
        ]

        # call function and assert result
        resp = client.multicall2(calls)
        assert resp == expected_params

    @responses.activate
    def test_insert_secret_with_multicall2(self):
        # create client with secret
        secret = "hello"
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params_1 = ["0000000000000001"]
        params_2 = ["2fa07b6e85c40205"]
        calls = [(client.REMOVE, params_1), (client.REMOVE, params_2)]
        # copy params and insert secret
        expected_params = [
            [
                {"methodName": client.REMOVE, "params": deepcopy(params_1)},
                {"methodName": client.REMOVE, "params": deepcopy(params_2)},
            ]
        ]
        for param in expected_params[0]:
            param["params"].insert(0, f"token:{secret}")

        # call function and assert result
        resp = client.multicall2(calls, insert_secret=True)
        assert resp == expected_params


class TestClientExceptionClass:
    @responses.activate
    def test_call_raises_custom_error(self):
        client = Client()
        responses.add(
            responses.POST, client.server, json={"error": {"code": 1, "message": "Custom message"}}, status=200
        )
        with pytest.raises(ClientException, match=r"Custom message") as e:
            client.call("aria2.method")
            assert e.code == 1

    @responses.activate
    def test_call_raises_known_error(self):
        client = Client()
        responses.add(
            responses.POST,
            client.server,
            json={"error": {"code": JSONRPC_PARSER_ERROR, "message": "Custom message"}},
            status=200,
        )
        with pytest.raises(ClientException, match=rf"{JSONRPC_CODES[JSONRPC_PARSER_ERROR]}\nCustom message") as e:
            client.call("aria2.method")
            assert e.code == JSONRPC_PARSER_ERROR


class TestClientClass:
    def test_add_metalink_method(self, server):
        # get file contents
        with open(DEBIAN_METALINK, "rb") as stream:
            metalink_contents = stream.read()
        encoded_contents = b64encode(metalink_contents).decode("utf-8")

        assert server.client.add_metalink(encoded_contents)

    def test_add_torrent_method(self, server):
        # get file contents
        with open(BUNSENLABS_TORRENT, "rb") as stream:
            torrent_contents = stream.read()
        encoded_contents = b64encode(torrent_contents).decode("utf-8")

        assert server.client.add_torrent(encoded_contents, [])

    def test_add_uri_method(self, server):
        assert server.client.add_uri([BUNSENLABS_MAGNET])
        assert server.client.add_uri(XUBUNTU_MIRRORS)

    def test_global_option_methods(self, tmp_path, port):
        with Aria2Server(tmp_path, port, config=CONFIGS_DIR / "max-5-dls.conf") as server:
            max_concurrent_downloads = server.client.get_global_option()["max-concurrent-downloads"]
            assert max_concurrent_downloads == "5"

            assert server.client.change_global_option({"max-concurrent-downloads": "10"}) == "OK"

            max_concurrent_downloads = server.client.get_global_option()["max-concurrent-downloads"]
            assert max_concurrent_downloads == "10"

    def test_option_methods(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="max-dl-limit-10000.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            max_download_limit = server.client.get_option(gid=gid)["max-download-limit"]
            assert max_download_limit == "10000"

            assert server.client.change_option(gid, {"max-download-limit": "20000"}) == "OK"

            max_download_limit = server.client.get_option(gid)["max-download-limit"]
            assert max_download_limit == "20000"

    def test_position_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
            gids = server.client.tell_waiting(0, 5, keys=["gid"])
            first, second = [r["gid"] for r in gids]
            assert server.client.change_position(second, 0, "POS_SET") == 0
            assert server.client.change_position(second, 5, "POS_CUR") == 1

    def test_change_uri_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="1-dl-2-uris.txt") as server:
            gid = server.client.tell_waiting(0, 1, keys=["gid"])[0]["gid"]
            assert server.client.change_uri(gid, 1, ["http://localhost:8779/1024"], ["http://localhost:8779/1k"]) == [
                1,
                1,
            ]
            assert server.client.change_uri(gid, 1, ["http://localhost:8779/1k"], []) == [1, 0]

    def test_force_pause_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="big-download.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.force_pause(gid) == gid

    def test_force_pause_all_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="2-dls.txt") as server:
            assert server.client.force_pause_all() == "OK"

    def test_force_remove_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="big-download.txt") as server:
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.force_remove(gid)
            assert server.client.tell_status(gid, keys=["status"])["status"] == "removed"

    def test_force_shutdown_method(self, server):
        assert server.client.force_shutdown() == "OK"
        with pytest.raises(requests.ConnectionError):
            for retry in range(10):
                server.client.list_methods()
                time.sleep(1)

    def test_get_files_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
            gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            assert len(server.client.get_files(gid)) == 1

    def test_get_global_stat_method(self, server):
        assert server.client.get_global_stat()

    def test_get_peers_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="max-dl-limit-10000.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert not server.client.get_peers(gid)

    def test_get_servers_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="max-dl-limit-10000.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.get_servers(gid)

    def test_get_session_info_method(self, server):
        assert server.client.get_session_info()

    def test_get_uris_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="1-dl-2-uris.txt") as server:
            gid = server.client.tell_waiting(0, 1, keys=["gid"])[0]["gid"]
            assert server.client.get_uris(gid) == [
                {"status": "waiting", "uri": "http://localhost:8779/1024"},
                {"status": "waiting", "uri": "http://localhost:8779/1k"},
            ]

    def test_get_version_method(self, server):
        assert server.client.get_version()

    def test_list_methods_method(self, server):
        assert server.client.list_methods()

    def test_list_notifications_method(self, server):
        assert server.client.list_notifications()

    def test_multicall_method(self, server):
        assert server.client.multicall(
            [[{"methodName": server.client.LIST_METHODS}, {"methodName": server.client.LIST_NOTIFICATIONS}]]
        )

    def test_multicall2_method(self, server):
        assert server.client.multicall2([(server.client.LIST_METHODS, []), (server.client.LIST_NOTIFICATIONS, [])])

    def test_pause_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.pause(gid) == gid

    def test_pause_all_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="2-dls.txt") as server:
            assert server.client.pause_all() == "OK"

    def test_purge_download_result_method(self, server):
        assert server.client.purge_download_result() == "OK"

    def test_remove_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.remove(gid)
            assert server.client.tell_status(gid, keys=["status"])["status"] == "removed"

    def test_remove_download_result_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            server.client.remove(gid)
            assert server.client.remove_download_result(gid) == "OK"
            assert len(server.client.tell_stopped(0, 1)) == 0

    def test_save_session_method(self, tmp_path, port):
        session_input = SESSIONS_DIR / "1-dl.txt"
        with Aria2Server(tmp_path, port, session=session_input) as server:
            session_output = server.tmp_dir / "_session.txt"
            server.client.change_global_option({"save-session": str(session_output)})
            assert server.client.save_session() == "OK"
            with open(session_input) as stream:
                input_contents = stream.read()
            with open(session_output) as stream:
                output_contents = stream.read()
            for line in input_contents.split("\n"):
                assert line in output_contents

    def test_shutdown_method(self, server):
        assert server.client.shutdown() == "OK"
        with pytest.raises(requests.ConnectionError):
            for retry in range(10):
                server.client.list_methods()
                time.sleep(1)

    def test_tell_active_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="big-download.txt") as server:
            time.sleep(0.1)
            if server.api.get_download("0000000000000001").has_failed:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert len(server.client.tell_active(keys=["gid"])) > 0

    def test_tell_status_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
            gid = server.client.tell_waiting(0, 1, keys=["gid"])[0]["gid"]
            assert server.client.tell_status(gid)

    def test_tell_stopped_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="very-small-download.txt") as server:
            download = server.api.get_download("0000000000000001")
            while not download.live.is_complete:
                if download.has_failed:
                    pytest.xfail("Failed to establish connection (sporadic error)")
                time.sleep(0.1)
            assert len(server.client.tell_stopped(0, 1, keys=["gid"])) > 0

    def test_tell_waiting_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
            assert server.client.tell_waiting(0, 5, keys=["gid"]) == [
                {"gid": "0000000000000001"},
                {"gid": "0000000000000002"},
            ]

    def test_unpause_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
            gid = server.client.tell_waiting(0, 1, keys=["gid"])[0]["gid"]
            assert server.client.unpause(gid) == gid

    def test_unpause_all_method(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
            assert server.client.unpause_all() == "OK"

    def test_listen_to_notifications_no_server(self):
        client = Client(port=7035)
        client.listen_to_notifications(timeout=1)

    def test_listen_to_notifications_no_callbacks(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

            def thread_target():
                server.client.listen_to_notifications(timeout=1, handle_signals=False)

            thread = threading.Thread(target=thread_target)
            thread.start()
            server.client.unpause("0000000000000001")
            time.sleep(3)
        thread.join()

    def test_listen_to_notifications_callbacks(self, tmp_path, port, capsys):
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

            def thread_target():
                server.client.listen_to_notifications(
                    on_download_start=lambda gid: print("started " + gid), timeout=1, handle_signals=False
                )

            thread = threading.Thread(target=thread_target)
            thread.start()
            time.sleep(1)
            server.client.unpause("0000000000000001")
            time.sleep(3)
        thread.join()
        assert capsys.readouterr().out == "started 0000000000000001\n"

    def test_listen_to_notifications_then_stop(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

            def thread_target():
                server.client.listen_to_notifications(timeout=1, handle_signals=False)

            thread = threading.Thread(target=thread_target)
            thread.start()
            server.client.stop_listening()
            thread.join()

    def test_listen_to_notifications_then_stop_with_signal(self, tmp_path, port):
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

            def thread_target():
                time.sleep(2)
                os.kill(os.getpid(), signal.SIGTERM)

            thread = threading.Thread(target=thread_target)
            thread.start()
            server.client.listen_to_notifications(timeout=1, handle_signals=True)
            thread.join()


class TestNotificationClass:
    def test_init(self):
        notification = Notification("random", "random")
        assert notification

    def test_get(self):
        message = {"method": "random_event", "params": [{"gid": "random_gid"}]}
        assert Notification.get_or_raise(message)

    def test_raise(self):
        message = {"error": {"code": 9000, "message": "it's over 9000"}}
        with pytest.raises(ClientException):
            Notification.get_or_raise(message)


class TestSecretToken:
    def test_works_correctly_with_secret_set(self, tmp_path, port):
        with Aria2Server(tmp_path, port, secret="this secret token") as server:
            assert server.client.get_version()

    def test_does_not_authorize_with_invalid_secret(self, tmp_path, port):
        with Aria2Server(tmp_path, port, secret="this secret token") as server:
            server.client.secret = "invalid secret token"
            with pytest.raises(ClientException):
                server.client.get_version()



"""Generic command to add downloads."""

import sys
from typing import List

from loguru import logger

from aria2p.api import API


def add(
    api: API,
    uris: List[str] = None,
    from_file: str = None,
    options: dict = None,
    position: int = None,
) -> int:
    """
    Add magnet subcommand.

    Arguments:
        api: The API instance to use.
        uris: The URIs or file-paths to add.
        from_file: Path to the file to read uris from.
            Deprecated: Every URI that is a valid file-path
            and is not a torrent or a metalink is now read as an input file.
        options: String of aria2c options to add to download.
        position: Position to add new download in the queue.

    Returns:
        int: 0 if OK else 1.
    """
    uris = uris or []

    if from_file:
        logger.warning(
            "Deprecation warning: every URI that is a valid file-path "
            "and is not a torrent or a metalink is now read as an input file.",
        )

    new_downloads = []

    for uri in uris:
        created_downloads = api.add(uri, options=options, position=position)
        new_downloads.extend(created_downloads)
        if position is not None:
            position += len(created_downloads)

    if new_downloads:
        for new_download in new_downloads:
            print(f"Created download {new_download.gid}")
        return 0

    print("No new download was created", file=sys.stderr)
    return 1



"""Command to run the text user-interface."""

import sys

from aria2p.api import API

try:
    from aria2p.interface import Interface
except ImportError:
    Interface = None  # type: ignore  # noqa: WPS440 (variable overlap)


def top(api: API) -> int:
    """
    Top subcommand.

    Arguments:
        api: The API instance to use.

    Returns:
        int: Always 0.
    """
    if Interface is None:
        print(
            "The top-interface dependencies are not installed. Try running `pip install aria2p[tui]` to install them.",
            file=sys.stderr,
        )
        return 1

    interface = Interface(api)
    success = interface.run()
    return 0 if success else 1



"""Command to resume downloads."""

import sys
from typing import List

from aria2p.api import API
from aria2p.client import ClientException


def resume(api: API, gids: List[str] = None, do_all: bool = False) -> int:
    """
    Resume subcommand.

    Arguments:
        api: The API instance to use.
        gids: The GIDs of the downloads to resume.
        do_all: Pause all downloads if True.

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if do_all:
        if api.resume_all():
            return 0
        return 1

    try:
        downloads = api.get_downloads(gids)
    except ClientException as error:
        print(str(error), file=sys.stderr)
        return 1

    result = api.resume(downloads)

    if all(result):
        return 0

    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)

    return 1



# Why does this file exist, and why not put this in `__main__`?
#
# You might be tempted to import things from `__main__` later,
# but that will cause problems: the code will get executed twice:
#
# - When you run `python -m aria2p` python will execute
#   `__main__.py` as a script. That means there won't be any
#   `aria2p.__main__` in `sys.modules`.
# - When you import `__main__` it will get executed again (as a module) because
#   there's no `aria2p.__main__` in `sys.modules`.

"""Module that contains the command line application."""

import argparse

from aria2p.client import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_TIMEOUT


def check_args(parser: argparse.ArgumentParser, opts: argparse.Namespace) -> None:  # noqa: WPS231 (complex)
    """
    Additional checks for command line arguments.

    Arguments:
        parser: An argument parser.
        opts: Parsed options.
    """
    subparsers = [
        action
        for action in parser._actions  # noqa: WPS437 (protected attribute)
        if isinstance(action, argparse._SubParsersAction)  # noqa: WPS437
    ][0].choices

    gid_commands = (
        "pause",
        "stop",
        "remove",
        "rm",
        "del",
        "delete",
        "resume",
        "start",
        "autopurge",
        "autoclear",
        "autoremove",
    )

    if opts.subcommand in gid_commands:
        if not opts.do_all and not opts.gids:
            subparsers[opts.subcommand].error("the following arguments are required: gids or --all")
        elif opts.do_all and opts.gids:
            subparsers[opts.subcommand].error("argument -a/--all: not allowed with arguments gids")
    elif opts.subcommand:
        if opts.subcommand in {"add", "add-magnet", "add-magnets"}:
            if not opts.uris and not opts.from_file:
                subparsers[opts.subcommand].error("the following arguments are required: uris")
        elif opts.subcommand.startswith("add-torrent"):
            if not opts.torrent_files and not opts.from_file:
                subparsers[opts.subcommand].error("the following arguments are required: torrent_files")
        elif opts.subcommand.startswith("add-metalink"):
            if not opts.metalink_files and not opts.from_file:
                subparsers[opts.subcommand].error("the following arguments are required: metalink_files")


def parse_options_string(options_string: str = None) -> dict:
    """
    Parse string of options.

    Arguments:
        options_string: String of aria2c options.

    Raises:
        ArgumentTypeError: When the options string is not correctly formatted.

    Returns:
        Dictionary containing aria2c options.
    """
    options_string = options_string or ""
    options = {}
    for download_option in options_string.split(";"):
        try:
            opt, val = download_option.split("=", 1)
        except ValueError:
            raise argparse.ArgumentTypeError(
                "Options strings must follow this format:\nopt-name=opt-value;opt-name2=opt-value2"
            )
        options[opt.strip()] = val.strip()
    return options


def get_parser() -> argparse.ArgumentParser:
    """
    Return a parser for the command-line options and arguments.

    Returns:
        An argument parser.
    """
    usage = "%(prog)s [GLOBAL_OPTS...] COMMAND [COMMAND_OPTS...]"  # noqa: WPS323 (%-formatting)
    description = "Command-line tool and Python library to interact with an `aria2c` daemon process through JSON-RPC."
    parser = argparse.ArgumentParser(add_help=False, usage=usage, description=description, prog="aria2p")

    main_help = "Show this help message and exit. Commands also accept the -h/--help option."
    subcommand_help = "Show this help message and exit."

    global_options = parser.add_argument_group(title="Global options")
    global_options.add_argument("-h", "--help", action="help", help=main_help)

    global_options.add_argument(
        "-p",
        "--port",
        dest="port",
        default=DEFAULT_PORT,
        type=int,
        help="Port to use to connect to the remote server.",
    )
    global_options.add_argument(
        "-H",
        "--host",
        dest="host",
        default=DEFAULT_HOST,
        help="Host address for the remote server.",
    )
    global_options.add_argument(
        "-s",
        "--secret",
        dest="secret",
        default="",
        help="Secret token to use to connect to the remote server.",
    )
    global_options.add_argument(
        "-L",
        "--log-level",
        dest="log_level",
        default=None,
        help="Log level to use",
        choices=("TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"),
        type=str.upper,
    )
    global_options.add_argument(
        "-P",
        "--log-path",
        dest="log_path",
        default=None,
        help="Log path to use. Can be a directory or a file.",
    )
    global_options.add_argument(
        "-T",
        "--client-timeout",
        dest="client_timeout",
        default=DEFAULT_TIMEOUT,
        type=float,
        help=f"Timeout in seconds for requests to the remote server. Floats supported. Default: {DEFAULT_TIMEOUT}.",
    )

    # ========= SUBPARSERS ========= #
    subparsers = parser.add_subparsers(dest="subcommand", title="Commands", metavar="", prog="aria2p")

    def subparser(command: str, text: str, **kwargs) -> argparse.ArgumentParser:  # noqa: WPS430 (nested function)
        sub = subparsers.add_parser(command, add_help=False, help=text, description=text, **kwargs)
        sub.add_argument("-h", "--help", action="help", help=subcommand_help)
        return sub

    add_parser = subparser("add", "Add downloads with URIs/Magnets/torrents/Metalinks.")
    add_magnets_parser = subparser("add-magnets", "Add downloads with Magnet URIs.", aliases=["add-magnet"])
    add_metalinks_parser = subparser("add-metalinks", "Add downloads with Metalink files.", aliases=["add-metalink"])
    add_torrents_parser = subparser("add-torrents", "Add downloads with torrent files.", aliases=["add-torrent"])
    subparser(
        "purge",
        "Automatically purge completed/removed/failed downloads.",
        aliases=["autoclear", "autopurge", "autoremove"],
    )
    call_parser = subparser("call", "Call a remote method through the JSON-RPC client.")
    pause_parser = subparser("pause", "Pause downloads.", aliases=["stop"])
    remove_parser = subparser("remove", "Remove downloads.", aliases=["rm", "del", "delete"])
    resume_parser = subparser("resume", "Resume downloads.", aliases=["start"])
    subparser("show", "Show the download progression.")
    subparser("top", "Launch the top-like interactive interface.")
    listen_parser = subparser("listen", "Listen to notifications.")

    # ========= REUSABLE OPTIONS ========= #
    def add_options_argument(_parser):  # noqa: WPS430
        _parser.add_argument(
            "-o",
            "--options",
            dest="options",
            type=parse_options_string,
            help="Options for the new download(s), separated by semi-colons. "
            "Example: 'dir=~/aria2_downloads;max-download-limit=100K'",
        )

    def add_position_argument(_parser):  # noqa: WPS430
        _parser.add_argument(
            "-p",
            "--position",
            dest="position",
            type=int,
            help="Position at which to add the new download(s) in the queue. Starts at 0 (top).",
        )

    # ========= CALL PARSER ========= #
    call_parser.add_argument(
        "method",
        help=(
            "The method to call (case insensitive). "
            "Dashes and underscores will be removed so you can use as many as you want, or none. "
            "Prefixes like 'aria2.' or 'system.' are also optional."
        ),
    )
    call_parser_mxg = call_parser.add_mutually_exclusive_group()
    call_parser_mxg.add_argument(
        "-P",
        "--params-list",
        dest="params",
        nargs="+",
        help="Parameters as a list of strings.",
    )
    call_parser_mxg.add_argument(
        "-J",
        "--json-params",
        dest="params",
        help="Parameters as a JSON string. You should always wrap it at least once in an array '[]'.",
    )

    # ========= ADD PARSER ========= #
    add_parser.add_argument("uris", nargs="*", help="The URIs/file-paths to add.")
    add_parser.add_argument("-f", "--from-file", dest="from_file", help="Load URIs from a file.")
    add_options_argument(add_parser)
    add_position_argument(add_parser)

    # ========= ADD MAGNET PARSER ========= #
    add_magnets_parser.add_argument("uris", nargs="*", help="The magnet URIs to add.")
    add_magnets_parser.add_argument("-f", "--from-file", dest="from_file", help="Load URIs from a file.")
    add_options_argument(add_magnets_parser)
    add_position_argument(add_magnets_parser)

    # ========= ADD TORRENT PARSER ========= #
    add_torrents_parser.add_argument("torrent_files", nargs="*", help="The paths to the torrent files.")
    add_torrents_parser.add_argument("-f", "--from-file", dest="from_file", help="Load file paths from a file.")
    add_options_argument(add_torrents_parser)
    add_position_argument(add_torrents_parser)

    # ========= ADD METALINK PARSER ========= #
    add_metalinks_parser.add_argument("metalink_files", nargs="*", help="The paths to the metalink files.")
    add_metalinks_parser.add_argument("-f", "--from-file", dest="from_file", help="Load file paths from a file.")
    add_options_argument(add_metalinks_parser)
    add_position_argument(add_metalinks_parser)

    # ========= PAUSE PARSER ========= #
    pause_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to pause.")
    pause_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Pause all the downloads.")
    pause_parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        help="Pause without contacting servers first.",
    )

    # ========= RESUME PARSER ========= #
    resume_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to resume.")
    resume_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Resume all the downloads.")

    # ========= REMOVE PARSER ========= #
    remove_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to remove.")
    remove_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Remove all the downloads.")
    remove_parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        help="Remove without contacting servers first.",
    )

    # ========= LISTEN PARSER ========= #
    listen_parser.add_argument(
        "-c",
        "--callbacks-module",
        dest="callbacks_module",
        help="Path to the Python module defining your notifications callbacks.",
    )
    listen_parser.add_argument(
        "event_types",
        nargs="*",
        help="The types of notifications to process: "
        "start, pause, stop, error, complete or btcomplete. "
        "Example: aria2p listen error btcomplete. "
        "Useful if you want to spawn multiple specialized aria2p listener, "
        "for example one for each type of notification, "
        "but still want to use only one callback file.",
    )
    listen_parser.add_argument(
        "-t",
        "--timeout",
        dest="timeout",
        type=float,
        default=5,
        help="Timeout in seconds to use when waiting for data over the WebSocket at each iteration. "
        "Use small values for faster reactivity when stopping to listen.",
    )

    return parser
